//  Copyright (c) 1997-2000 by Mark F. Adams 
//  $Id: geo_coar_fix.C,v 1.2 2004/06/16 00:12:26 adams Exp $
//  Author:             Mark F. Adams
//  Version:            2.0
//  Filename:           geo_coar_fix.C
//
#include "prom_grid.hh"

static PromList<PromStack*> *bf_list = NULL; // local buffer

#undef __FUNCT__  
#define __FUNCT__ "PromGeoGrid::FixMesh"
/* PromGeoGrid::FixMesh() ******************************************
 *
 *  - take block of elems (w/ fine nodes in these coarse elems),
 *    delete elements, cover nodes.
 *
 *  INPUT:
 *  SIDE EFFECTS:
 *    - deletes  extraneous elements
 *
 *  RETURN:
 *    - Petsc error code
 */
int PromGeoGrid::FixMesh( PromList<PromPoly*> *elemList )  
{
  int       xx, ierr, numelems0, numelems1,nlost=0,nc,mype=comm_.mype();
  int       loops,level=getLevel(),verb=options_.verbose_>1; bool flag;
  gfloat    maxoff;
  PromGrid *lastGrid = prevgrid_; 
  MPI_Comm  comm = (verb>1||comm_.np()<=1) ? PETSC_COMM_SELF:lastGrid->gComm_;
  if( !bf_list ){bf_list=new PromList<PromStack*>(10);CHKNEW_void(bf_list,1);}
  
  // get option fpr 
  ierr = options_.GetOptionInt("-geomg_move", &loops, &flag);  CHKERRQ(ierr);
  if( !flag ) loops = 0;
  
  // delete envelope elements
  if(verb)PetscPrintf(comm,"[%d]Remove env elems(%d)\n", mype,level );
  numelems0 = getNumTets();
  ierr = RemoveEnvElems( nlost, elemList );                   CHKERRQ(ierr);
  numelems1 = getNumTets();
  //PetscPrintf( nlost==0 ? comm : PETSC_COMM_SELF,
  if(verb)PetscPrintf(comm,"[%d]--done,deleted %d/%d env elems,lost %d\n",
		      mype, numelems0 - numelems1, numelems0, nlost);
  
  // delete unconnected elems if enough nodes
  if(verb)PetscPrintf( comm, "[%d]Remove unconnected part elems(%d)\n",
		       mype,level);
  numelems0 = numelems1;
  ierr = RemoveDisconnectedPartElems( nlost, elemList );       CHKERRQ(ierr);
  numelems1 = getNumTets();
  if(verb)PetscPrintf( comm, "[%d]--done, deleted %d/%d elems, lost %d\n",
		       mype, numelems0 - numelems1, numelems0, nlost);
  
  // other deletes
  if(verb)PetscPrintf(comm,"[%d]Remove unconnected elems(%d)\n",mype,level);
  numelems0 = numelems1; 
  ierr = RemoveUnconnectedElems_easy( elemList );            CHKERRQ(ierr);
  numelems1 = getNumTets();
  if(verb)PetscPrintf(comm,"[%d]--done, deleted %d/%d uc elems,lost %d\n",
		      mype, numelems0 - numelems1, numelems0, nlost); 

  if( isActive() && loops > 0 ) {
    // move nodes to cover misssed nodes, and resolve all nodes
    if(verb)PetscPrintf( comm,"[%d]MoveNodes(%d) - %d loops, %d todo\n",
			     mype, level, loops, nlost );
    //loops = loops > 0 ?  loops + (getLevel()-1)*2 : 0;
    ierr = MoveNodes( nlost, loops, maxoff, elemList );       CHKERRQ(ierr);
    if(verb) PetscPrintf( comm,"[%d]MoveNodes--done\n", mype);
  }
  if( !elemList->isEmpty() ) {
    // put all uncovered nodes some place, sets elem pointer
    ierr = ResolveNodes( nlost, maxoff, elemList );           CHKERRQ(ierr);
    // clear elem lists of nodes (to clear nodes child lists)
    Position pos = elemList->getHeadPosition();
    while ( pos ) { 
      PromDTet *elem = dynamic_cast<PromDTet*>(elemList->getNext(pos));
      elem->ClearChildren(); 
    }
  }
  else if( nlost != 0 ) SETERRQ1(1,"geometric algo. failure, lost %d",nlost);

  // reclassify nodes (with these new elements)
  if( getLevel() > 1 ){  
    if(verb) PetscPrintf(comm,"\t[%d]SetNodeTypes first\n",mype); 
    ierr = SetNodeTypes( );                          CHKERRQ(ierr);
  }  
  else if( isActive() ) {
    // see if too many corners
    for( xx = nc = 0 ; xx < nLocalNd_ ; xx++ ){
      PromNode *curr = nodes_[xx];
      if( curr->getTopoType() == PROM_CORNER ) nc++;
    }
    ierr = MPI_Allreduce(&nc,&xx,1,MPI_INT,MPI_SUM,gComm_); CHKERRQ(ierr); 
    if( (100*xx)/nglobal() > 5 ) {
      if(verb)PetscPrintf(comm,"\t[%d]SetNodeType first(%d/%d corner)\n",
			      mype,xx,nglobal()); 
      ierr = SetNodeTypes();                          CHKERRQ(ierr);
    }
  }
  // clean up marks
  Position pos2, posx = elemList->getHeadPosition();
  while ( (pos2=posx) ) {
    PromDTet *elem = dynamic_cast<PromDTet*>(elemList->getNext(posx));    
    if( elem->isDead() ) elemList->removeAt(pos2);
    else elem->mark(FALSE);
  }
  //set last grids parents 
  for( xx = 0 ; xx < lastGrid->nNodes_ ; xx++ ){
    PromNode *fine = lastGrid->nodes_[xx];
    assert( !fine->isMarked() );
    assert( fine->nParents() || fine->isGhost() );
    for( int tt = 0 ; tt < fine->nParents() ; tt++ ){
      int proc = fine->parents_->data_[tt]->proc();
      fine->parents_->data_[tt]->setProc( masterProc(proc) );
      assert( !fine->parents_->data_[tt]->isMarked() );
    }
  } 
  // clean up
  delete bf_list; bf_list = NULL;
 
  return 0;
}
 
#undef __FUNCT__  
#define __FUNCT__ "PromGeoGrid::RemoveEnvElems"
/* PromGeoGrid::RemoveEnvElems ******************************************
 *
 *  - deletes envelope elems
 *
 *  INPUT:
 *    nlost: number fine nodes w/o elems, in this nodes 'children'.(in/out)
 *
 *  SIDE EFFECTS:
 *    - deletes envelope elements
 *    - adds orphaned nodes to interior (non-envelope) node.list
 *
 *  RETURN:
 *    - Petsc error code
 */
int PromGeoGrid::RemoveEnvElems( int &nlost, PromList<PromPoly*> *elemList )
{
  PromDTet  *elem;
  int    i;      

  // loop over  envelope nodes and remove its elements
  for( i = 0 ; i < PROM_NUMENV ; i++ ) {
    PromGeoNode *envnode = env_[i];    
    Position pos = envnode->symbData_->elemList.getHeadPosition(); 
    while( pos ) {  
      elem=dynamic_cast<PromDTet*>(envnode->symbData_->elemList.getNext(pos));
      if( elem->isDead() ) continue; // deleted
      evacuateChildren( elem );     // evacuate nodes
    }
  } 
  // loop over envelope nodes and remove its elements
  for( i = PROM_NUMENV-1 ; i >=0 ; i-- ) {
    PromNode *envnode = env_[i];    
    while( !envnode->symbData_->elemList.isEmpty()  ) {  
      elem=dynamic_cast<PromDTet*>(envnode->symbData_->elemList.removeHead());
      if( elem->isDead() ) continue; // deleted
      int n = retireElement( elem ); 
      nlost += n; // hard delete
    }  
  }    
  
  // clean up
  Position pos2, posx = elemList->getHeadPosition();
  while ( (pos2 = posx) != NULL ) {
    elem = dynamic_cast<PromDTet*>(elemList->getNext(posx));    
    if( elem->isDead() ) elemList->removeAt(pos2);
  }

  return 0;
}

#undef __FUNCT__  
#define __FUNCT__ "PromGeoGrid::RemoveOutsideSlivers"
/* PromGeoGrid::RemoveOutsideSlivers() ***************************************
 *
 *  - deletes slivers on outside surfaces
 *
 *  INPUT:
 *    nlost: number fine nodes w/o elems, in this nodes 'children'.(in/out)
 *
 *  SIDE EFFECTS:
 *  RETURN:
 *    - PETSc error flag
 */
int PromGeoGrid::RemoveOutsideSlivers(int &nlost,PromList<PromPoly*> *elemList)
{
  gfloat tol = 1.e-5;
  int    ierr;
  Position pos2, posx;
  
  do {  
    // delete slivers on outside - and remove dead nodes from 'elemList'
    posx = elemList->getHeadPosition();
    while ( (pos2 = posx) != NULL ) {
      PromDTet *elem = dynamic_cast<PromDTet*>(elemList->getNext(posx));    
      if ( elem->isDead() ) {
        elemList->removeAt(pos2);
        continue;
      }
      if(elem->willStrandNode(ierr) || ierr || elem->numGhosts() == 4)continue;
      
      // get rid of slivers on surface
      bool kill = (elem->numExposedFaces() >= 2 && elem->quality() < tol );
      if( kill && evacuateChildren( elem ) == 0 ) {
	int n = retireElement( elem ); 
	assert( n == 0 ); 
	elemList->removeAt(pos2);
      } 
    }
  }while( (tol *= 10.0) < 0.5 ); 
  
  return 0;
}

#undef __FUNCT__  
#define __FUNCT__ "PromGeoGrid::RemoveOutsides"
/* PromGeoGrid::RemoveOutsides() ***************************************
 *
 *  - deletes
 *
 *  INPUT:
 *    nlost: number fine nodes w/o elems, in this nodes 'children'.(in/out)
 *
 *  SIDE EFFECTS:
 *  RETURN:
 *    - PETSc error flag
 */
int PromGeoGrid::RemoveOutsides( int &, PromList<PromPoly*> *elemList )
{
  int    ierr, k, kill, ii, kk, nkilled;
  Position pos2, posx;
    
  for( int xxx = 0 ; xxx < 5 ; xxx++ ){
    nkilled = 0;
    // delete slivers on outside - and remove dead nodes from 'elemList'
    posx = elemList->getHeadPosition();
    while ( (pos2 = posx) != NULL ) {
      PromDTet *elem = dynamic_cast<PromDTet*>(elemList->getNext(posx));    
      if ( elem->isDead() ) {
	elemList->removeAt(pos2);
      }
      else if( elem->numExposedFaces() >= 1 && !elem->willStrandNode(ierr) 
	       && ierr == 0 ){
	for( kill = TRUE, k = 0 ; k < 4 && kill ; k++ ) { 
	  for( kk = ii = 0 ; kk < 4 ; kk++ ) { 
	    if( k != kk && elem->neigh_[kk] ) ii++;
	  }
	  if( ii == 3 && elem->nodes_[k]->isExterior() ) kill = FALSE; 
	}
	if( kill && evacuateChildren( elem ) == 0 && elem->numGhosts() < 4 ){ 
	  int n = retireElement( elem ); 
	  assert(n == 0); 
	  elemList->removeAt(pos2);
	  nkilled++;
	}
      }
    }
    if( nkilled == 0 ) break;
  }

  return 0;
}

#undef __FUNCT__  
#define __FUNCT__ "PromGeoGrid::RemoveDisconnectedPartElems"
/* PromGeoGrid::RemoveDisconnectedPartElems() ***************************
 *
 *  - remove elems from different parts
 *
 *  INPUT:
 *    nlost: number fine nodes w/o elems, in this nodes 'children'.(in/out)
 *
 *  SIDE EFFECTS:
 *  RETURN:
 *    - Petsc error code
 */
int PromGeoGrid::RemoveDisconnectedPartElems( int &nlost, 
					      PromList<PromPoly*> *elemList)
{
  int       ii, kk, k;
  Position  pos2, posx;

  // get rid of elems between diff parts
  posx = elemList->getHeadPosition();
  while ( (pos2 = posx) != NULL ) {
    PromDTet *elem = dynamic_cast<PromDTet*>(elemList->getNext(posx));  
    if( elem->isDead() ){ elemList->removeAt( pos2 ); continue; }
    for( kk = ii = 0 ; kk < 4 && ii == 0 ; kk++ ) { 
      for( k = kk+1 ; k < 4 && ii == 0 ; k++ ) { 
	if( elem->nodes_[kk]->color() != elem->nodes_[k]->color() ) ii++;
      }
    }
    if( ii ){
      evacuateChildren( elem ); 
      int n = retireElement( elem ); 
      nlost += n; 
      elemList->removeAt(pos2);
    }
  }
  
  return 0;
}

#undef __FUNCT__  
#define __FUNCT__ "PromGeoGrid::RemoveUnconnectedElems_easy"
/* PromGeoGrid::RemoveUnconnectedElems_easy() ******************************
 *
 *  - soft deletes unconnected elems
 *    * exterior nodes have been moved and marked
 *
 *  INPUT:
 *
 *  SIDE EFFECTS:
 *    - marks elements for possible deletion
 *
 *  RETURN:
 *    - Petsc error code
 */
int PromGeoGrid::RemoveUnconnectedElems_easy( PromList<PromPoly*> *elemList )
{
  // delete slivers on outside - and remove dead nodes from 'elemList'
  int       arr[4][4], isbad, k;
  Position  pos2, posx;
  PromTable<PromNode*> adjacs2(305);
  
  // get rid of easy very unconnected elems
  posx = elemList->getHeadPosition();
  while ( (pos2 = posx) != NULL ) {
    PromDTet *elem = dynamic_cast<PromDTet*>(elemList->getNext(posx));
    if( elem->isDead() ){ elemList->removeAt( pos2 ); continue; }

    // get conectivity (2 levels) array
    int maxdeg = 0, mindeg = 4;
    for( k = isbad = 0 ; k < 4 ; k++ ) { 
      PromNode *adji = elem->nodes_[k];
      if( adji->symbData_->elemList.getCount() == 1 ) { isbad = 1; break; }
      arr[k][k] = 1;
      if( k != 3 ) {
	assert( adjacs2.isEmpty() );
	for( int j = k+1 ; j < 4 ; j++ ) { 
	  PromNode *adjj = elem->nodes_[j]; // look for j through i
	  // check first layer
	  PromNode *found = adji->symbData_->adjac.find( adjj );
	  if( !found ) {                    // look deeper
	    TablePos tpos = adji->symbData_->adjac.getHeadPosition();
	    while ( tpos && !found ) {
	      PromNode *nn = adji->symbData_->adjac.getNext(tpos); 
	      if ( nn == adji ) continue;   
	      found = nn->symbData_->adjac.find( adjj );
	    } 
	  }
	  if( !found ) {   
	    if( adjacs2.isEmpty() ) {
	      // make second layer 
	      TablePos tpos = adji->symbData_->adjac.getHeadPosition();
	      while ( tpos ) {
		PromNode *n1 = adji->symbData_->adjac.getNext(tpos);
		if ( n1 == adji ) continue;   
		TablePos tpos2 = n1->symbData_->adjac.getHeadPosition( );
		while ( tpos2 ) {
		  PromNode *n2 = n1->symbData_->adjac.getNext(tpos2);
		  if( n2 != adji && !adji->symbData_->adjac.find(n2) ) {
		    adjacs2.add( n2 );
		  }
		} 
	      }
	    }
	    found = adjacs2.find( adjj ); assert(!found);
	    TablePos tpos = adjacs2.getHeadPosition();
	    while ( tpos && !found ) {
	      PromNode *n3 = adjacs2.getNext(tpos); 
	      found = n3->symbData_->adjac.find( adjj );
	      assert( n3 != adji && n3 != adjj ); 
	    } 
	  }
	  if ( found ) arr[k][j] = arr[j][k] = 1;
	  else         arr[k][j] = arr[j][k] = 0;
	}
	if( !adjacs2.isEmpty() ) adjacs2.removeAll();
      }
      int deg = arr[k][0] + arr[k][1] +  arr[k][2] +  arr[k][3];
      if ( deg > maxdeg ) maxdeg = deg;
      if ( deg < mindeg ) mindeg = deg;
    }
    if( ( maxdeg <= 2 || mindeg <= 1) && isbad == 0 && elem->numGhosts() < 4 ){
      // try to kill this
      const int numleft = evacuateChildren( elem );  
      if( numleft == 0 ) {
	bool killit = TRUE;
	for( k = 0 ; k < 4 ; k++ ) { 
	  PromNode *adji = elem->nodes_[k];
	  if( !adji->isEnvelope() && adji->symbData_->elemList.getCount() == 1
	      && !adji->isGhost() ) killit = FALSE;
	}
	if( killit && elem->numGhosts() < 4 ) {
	  int n = retireElement( elem ); assert( n == 0 );
	  //nlost += n;
	  elemList->removeAt(pos2); 
	}
      }
      else elem->mark(); // mark as unconnected
    }
  }

  return 0;
}

#undef __FUNCT__  
#define __FUNCT__ "PromGeoGrid::evacuateChildren"
/* PromGeoGrid::evacuateChildren ***********************************************
 *
 *    - Soft push to get rid of children, to (non-dead) nieghbors or
 *      add to a nodes child list
 *
 *  INPUT:
 *    - elem: element to evacuate
 *
 *  SIDE EFFECTS:
 *  RETURN:
 *    - number of children left
 */   
int PromGeoGrid::evacuateChildren( PromDTet *evacEl )
{ 
  assert(this!=NULL);  
  PromGeoNode*fine;
  Position    posx, pos, pos2;
  PromDTet   *elem, *newel;
  int         ii,mype=comm_.mype();  
  if( evacEl->isChildListEmpty() ) return 0;
  PromTable<PromDTet*> tetTab(57);

  // get children and try to put someplace else
  assert( bf_list->isEmpty() ); // list should be empty
  evacEl->removeChildren( bf_list );

  posx = bf_list->getHeadPosition();
  while ( (pos2 = posx) ) {
    fine = dynamic_cast<PromGeoNode*>(bf_list->getNext(posx));  
    if(fine->symbGeoData_->parElem != evacEl) 
      SETERRQ2(1,"fine->symbGeoData_->parElem(%p) != evacEl(%p)",
	       fine->symbGeoData_->parElem,evacEl);
    
    // try to get rid of 'fine'  
    newel = NULL;
    newel = fine->setElem(evacEl, parBank_, mype, 1.e-3, FALSE, 5, evacEl, FALSE);
    if( newel == NULL ) {
      newel = fine->setElem(evacEl, parBank_, mype, 1.e-2, FALSE, 10, evacEl, FALSE);
    }
    if( newel && newel->isSliver() ) newel = NULL;
    for( ii = 0 ; ii < 4 && newel == NULL ; ii++ ) {
      PromNode  *nn = evacEl->nodes_[ii]; assert(nn!=NULL);
      if( !nn->isEnvelope() ) {
	// look for common element - not very effiecent
	pos = nn->symbData_->elemList.getHeadPosition(); assert(pos!=NULL);
	while( pos && newel == NULL ) { 
	  elem = dynamic_cast<PromDTet*>(nn->symbData_->elemList.getNext(pos));
	  if( elem != evacEl && elem->numEnv() == 0 && !tetTab.find(elem) ) {
	    tetTab.add(elem);
	    newel = fine->setElem(elem,parBank_,mype,2.e-2,FALSE,5,evacEl,FALSE);
	    if( newel && newel->isSliver() ) newel = NULL;
	  }
	}
      }
    }
    tetTab.removeAll();
    // set element -- "better" elem - less envs and less ghosts
    if( newel ) {
      // see if this node is ok and new node has more ghosts
      bf_list->removeAt( pos2 );   // this is done
      elem = fine->setElem( newel, parBank_, mype,2.e-2, TRUE, 1, evacEl, FALSE );
      assert( fine->symbGeoData_->parElem == newel ); 
    }
    else assert( !fine->symbGeoData_->parElem->isDead() );
  } // 'bf_list'

  // put children back
  const int numleft = bf_list->getCount();
  while ( !bf_list->isEmpty() ) {
    fine = dynamic_cast<PromGeoNode*>(bf_list->removeHead());
    evacEl->addChild( fine );
    assert( fine->symbGeoData_->parElem == evacEl );
    assert( fine->nParents() > 0 ); 
  }

  return numleft;
}

// general "FixMesh" helpers
#undef __FUNCT__  
#define __FUNCT__ "PromGeoGrid::retireElement"
/* PromGeoGrid::retireElement *********************************
 *
 *  - Hard push of nodes out.
 *  - Prepair to get killed, clean up nodes of this element, kill it
 *
 *  INPUT:
 *    - deadElem: element to be resolved (deleted)
 *
 *  SIDE EFFECTS:
 *    - set nodes as "exterior"
 *    - kills the element
 *
 *  RETURN:
 *    -  number of moved nodes
 */
int PromGeoGrid::retireElement( PromDTet *deadElem )
{
  Position       pos2;
  PromGeoNode   *fine, *n1, *n2;
  int            ii,count=0,ierr,withEnv,nn,mype=comm_.mype();
  if( deadElem->isDead() )SETERRQ(1,"deadElem->isDead()");
  
  // get nodes inside of this exterior elem
  assert( bf_list->isEmpty() );
  deadElem->removeChildren( bf_list ); // get list to fix
  Kill( deadElem );         // kill element

  // remove adj to this killed element
  for ( ii = 0 ; ii < 4 ; ii++ ) {    
    // clear neighbor
    PromDTet  *elem = dynamic_cast<PromDTet*>(deadElem->neigh_[ii]); 
    if ( elem ) {
      assert( !elem->isDead() );
      if      (elem->neigh_[0] == deadElem) elem->neigh_[0] = NULL;
      else if (elem->neigh_[1] == deadElem) elem->neigh_[1] = NULL;
      else if (elem->neigh_[2] == deadElem) elem->neigh_[2] = NULL;
      else if (elem->neigh_[3] == deadElem) elem->neigh_[3] = NULL;
      else assert(0);
    }
  }
  // remove this from nodes 'elemList', fix any standed children
  for ( ii = 0 ; ii < 4 ; ii++ ) {
    n1 = dynamic_cast<PromGeoNode*>(deadElem->nodes_[ii]); 
    n1->setExterior(); // set exposed nodes as such
    if( !n1->isEnvelope() ) {
      PromList<PromPoly*> &list = n1->symbData_->elemList;
      pos2 = list.find( deadElem );   assert(pos2 != NULL);
      list.removeAt( pos2 );
      if( list.isEmpty() && n1->numChildren() > 0 ) {
	while( !n1->isChildListEmpty() ) {
	  n2 = dynamic_cast<PromGeoNode*>(n1->PopChild());
	  assert( n2->nParents() == 4 );
	  bf_list->addTail( n2 ); // could have other parElem
	}
      }
    }
    else assert( n1->isChildListEmpty() );
  }

  // loop over todo nodes 
  while( !bf_list->isEmpty() ) {
    fine = dynamic_cast<PromGeoNode*>(bf_list->removeHead());
    assert(fine->symbGeoData_->parElem!=NULL);
    if ( fine->nParents() == 1 ) continue; // direct child
    assert( fine->nParents() == 4 ); 
    assert(fine->getState()==PROM_DELETED || fine->getState()==PROM_SELECTED);
    PromNode *curr = NULL;
    if( fine->getState() == PROM_SELECTED ) {
      curr = fine->symbData_->parent; assert(curr!=NULL); 
    }
    else if( fine->getState() == PROM_DELETED ) {
      curr = fine->symbData_->parent->symbData_->parent; assert(curr!=NULL); 
    }
    else assert(0);
    const int color = curr->color();

    // add lost node to best candidates child list      (MOVE NODES)
    float maxshape = -1.e4; int maxi = -9, noffcolor = 0;
    for( ii = 0, n2 = NULL, withEnv = FALSE, nn = 4 ; ii < 4 ; ii++ ) {
      n1 = dynamic_cast<PromGeoNode*>(fine->symbGeoData_->parElem->nodes_[ii]);
      assert(n1!=NULL);
      assert(n1->color()>0 || n1->isEnvelope());
      if( n1->isEnvelope() ) { nn--; withEnv = TRUE; }
      else if( n1->color() != color ) noffcolor++;
      else if( !n1->symbData_->elemList.isEmpty() && !n1->isGhost() ){
	n2 = n1; 
	// node closest to this 
	shpfloat shape = fine->parents_->data_[ii]->getScalarP0();
	if( shape > maxshape  ){ 
	  maxshape = shape; maxi = ii; //want the lagest val (closest)
	}
      }
    }
    if( maxi != -9 ) {
      n2=dynamic_cast<PromGeoNode*>(fine->symbGeoData_->parElem->nodes_[maxi]);
    }
    if( (!withEnv && noffcolor==0) || n2 == NULL ){
      gfloat temp = 0.0;                 // set shapes here and now
      PromParent **parArr = fine->parents_->data_;
      for( ii = nn = 0 ; ii < 4 ; ii++ ){
	n1=dynamic_cast<PromGeoNode*>(fine->symbGeoData_->parElem->nodes_[ii]);
	if( !n1->isEnvelope() ) {
	  PromParent *pt = parArr[nn];
	  parArr[nn] = parArr[ii]; //pointer swap
	  parArr[ii] = pt;
	  temp += parArr[nn]->getScalarP0();
	  nn++;
	}
      }
      assert(nn>0); assert(temp!=0.0); 
      ierr = fine->SetnParents( parBank_, nn ); CHKERRQ(ierr); // reduction!!! 
      parArr = fine->parents_->data_;

      temp = 1./temp; // regularize
      for( ii = 0 ; ii < fine->nParents() ; ii++ ) {
	double val = parArr[ii]->getScalarP0();
	parArr[ii]->setScalarP0( val * temp );
      }
      if( nn != 4 && options_.verbose_>1 )
	PetscPrintf(options_.verbose_>2 ? PETSC_COMM_SELF : PETSC_COMM_WORLD,
		    "\t[%d]%s WARNING force %d (np=%d)\n",
		    mype,__FUNCT__,fine->index_(),fine->nParents());
      fine->symbGeoData_->parElem = NULL;
    }
    else { // withEnv and n2
      assert( !n2->symbData_->elemList.isEmpty() && !n2->isGhost());
      n2->addChild( fine ); // put lost node in 'list'
      count++; // ones to fix	  
    }
  }

  return count; 
}

#undef __FUNCT__  
#define __FUNCT__ "PromGeoGrid::ResolveNodes"
/* PromGeoGrid::ResolveNodes() ******************************************
 *
 *  - put fine nodes in this nodes "children" into some element
 *
 *  INPUT:
 *    - numUnfound: number of nodes to cover
 *    - maxoff: max off shape value. (out)
 *
 *  SIDE EFFECTS:
 *  RETURN:
 *    - PETSc error code.
 */
int PromGeoGrid::ResolveNodes( int numUnfound, gfloat &maxoff, 
			       PromList<PromPoly*> *elemList )
{
  int            xx,ierr,mype=comm_.mype(),vb=options_.verbose_>1,nf=numUnfound;
  PromNode      *curr;
  PromList<PromStack*> lostlist;
  MPI_Comm       comm = prevgrid_->gComm_;
  
  // get lost nodes, unmark
  for( xx = 0 ; xx < nNodes_ ; xx++ ){
    curr = nodes_[xx]; 
    curr->mark( FALSE );
    if( curr->getState() == PROM_NOTUSED){ 
      assert(curr->isChildListEmpty() ); continue;
    }
    assert( curr->isChildListEmpty() || !curr->isGhost() );
    curr->removeChildren ( &lostlist );  assert( curr->isChildListEmpty() ); 
  }
  if( numUnfound != lostlist.getCount() && vb ){
    PetscPrintf(PETSC_COMM_SELF, 
		"\t\t[%d]%s WARNING: %d unfound, %d lostlist. |elemlist|=%d\n",
		mype,__FUNCT__,numUnfound,lostlist.getCount(),
		elemList->getCount());
  }
  assert( numUnfound == lostlist.getCount() );

  // force unfound nodes
  if(numUnfound && vb>1)PetscPrintf(comm,"[%d]%s %d unfound,%d lostl\n",
				      mype,__FUNCT__,
				      numUnfound,lostlist.getCount());
  int maxi = -9;
  gfloat smax = 0.0, buff[4] = { 0., 0., 0., 0. };
  while ( !lostlist.isEmpty() ) {
    PromGeoNode* targ = dynamic_cast<PromGeoNode*>(lostlist.removeHead());    
    curr = targ->getParent(); assert(curr!=NULL); 
    targ->symbGeoData_->parElem = NULL;
    // get ~best elem
    PromDTet *alt, *elem = NULL, *best = NULL;
    gfloat fout, fbest=0.0;  
    // try to get a starter
    if( curr->symbData_->elemList.isEmpty() ) {
      TablePos tpos = curr->symbData_->adjac.getHeadPosition(); 
      while( tpos && curr->symbData_->elemList.isEmpty() ) {
	PromNode *nn = curr->symbData_->adjac.getNext(tpos); 
	if( !nn->symbData_->elemList.isEmpty() ) curr = nn;
      }
    }
    Position pos = curr->symbData_->elemList.getHeadPosition(); 
    while( pos && elem == NULL ){
      elem = dynamic_cast<PromDTet*>(curr->symbData_->elemList.getNext(pos));
      assert( targ->bufftable_ == NULL || targ->bufftable_->isEmpty() );
      elem = targ->setElem(elem, parBank_, mype, 2.e-1, FALSE, 2, NULL, FALSE,
			   &alt, &fout);
      if( alt && (best == NULL || (fout < fbest && !alt->isSliver())) ) {
	fbest = fout; best = alt;
      }
      if( elem && !elem->isSliver() ) break;
    }
    if( elem == NULL ) elem = best; 
    if( elem != NULL && !elem->isSliver() ){
      targ->setElem( elem, parBank_, mype, 2.e-1, TRUE, 1, NULL, FALSE );
    }
    if( targ->symbGeoData_->parElem == NULL ) {
      best = NULL;
      Position pos2, posx = elemList->getHeadPosition();
      while( (pos2=posx) && targ->symbGeoData_->parElem == NULL) {
	elem = dynamic_cast<PromDTet*>(elemList->getNext(posx));
	assert( elem != NULL );
	if( elem->isDead() ) elemList->removeAt(pos2);
	else {
	  elem = targ->setElem( elem, parBank_, mype, 3.e-1, FALSE, 3,
				NULL, FALSE, &alt, &fout );
	  if( alt && (best==NULL || (fout < fbest && !alt->isSliver())) ){
	    fbest = fout; best = alt;
	  }
	  if( elem && !elem->isSliver() ) { best = elem; break; }
	}
      }
      if(vb>1) {
	PetscPrintf(PETSC_COMM_SELF,"[%d]WARNING:%s, Deep look, minfout=%e\n",
		    mype, __FUNCT__,fbest);
      }
      assert( best != NULL );
      targ->setElem( best, parBank_, mype, 2.1*fbest, TRUE, 1, NULL, FALSE );
    }
    if(targ->symbGeoData_->parElem==NULL)SETERRQ(1,"Fail of setElem in MoveNodes");
    assert( !targ->symbGeoData_->parElem->isDead() );
    // get max off shape value for diagnostics
    ierr = targ->MaxOff( smax, &maxi, buff );                 CHKERRQ(ierr);

    numUnfound--;
  }
  assert( numUnfound == 0 );
  if(vb>0)
    PetscPrintf((smax == 0.0) ? comm : PETSC_COMM_SELF,
		"[%d]%s Max off %f: <%6.3f,%f,%6.3f,%6.3f> (fixed %4d)\n", 
		mype,__FUNCT__,smax, buff[0], buff[1], buff[2],
		buff[3],nf);
  // return 
  maxoff = smax;

  return 0;
}
